'''Utility functions for data kennel'''
from __future__ import print_function

import logging
import sys

from collections import defaultdict

logger = logging.getLogger(__name__)
YES_LIST = ['y', 't', 'yes', 'true', '1']


class EasyExit(Exception):
    """
    Raise this exception to exit your program with a log message and a non-zero status, but no stack trace
    (assuming you are running it with run_gracefully).
    """
    pass


def configure_logging(debug):
    '''Sets the data kennel logger to appropriate levels of chattiness.'''
    default_logger = logging.getLogger('')
    datadog_logger = logging.getLogger('datadog.api')
    requests_logger = logging.getLogger('requests')
    if debug:
        default_logger.setLevel(logging.DEBUG)
        datadog_logger.setLevel(logging.INFO)
        requests_logger.setLevel(logging.INFO)
    else:
        default_logger.setLevel(logging.INFO)
        datadog_logger.setLevel(logging.WARNING)
        requests_logger.setLevel(logging.WARNING)

    stream_handler = logging.StreamHandler(sys.__stdout__)
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s'))
    default_logger.addHandler(stream_handler)


def run_gracefully(main_function):
    """
    Run a "main" function with standardized exception trapping, to make it easy
    to avoid certain unnecessary stack traces.

    If debug logging is switched on, stack traces will return.
    """
    try:
        main_function()
    except EasyExit as msg:
        logger.error(str(msg))
        sys.exit(1)
    except KeyboardInterrupt:
        # swallow the exception unless we turned on debugging, in which case
        # we might want to know what infinite loop we were stuck in
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            raise
        sys.exit(1)


def print_table(rows, headers=None):
    """
    Convenience method for printing a list of dictionary objects into a table. Automatically sizes the
    columns to be the maximum size of any entry in the dictionary, and adds additional buffer whitespace.

    Params:
        rows -                  A list of dictionaries representing a table of information, where keys are the
                                headers of the table. Ex. { 'Name': 'John', 'Age': 23 }

        headers -               A list of the headers to print for the table. Must be a subset of the keys of
                                the dictionaries that compose the row. If a header isn't present or it's value
                                has a falsey value, the value printed is '-'.
    """
    columns_to_sizing = defaultdict(int)
    format_string = ''

    headers = headers or rows[0].keys()

    for row in rows:
        for header in headers:
            value = row.get(header, '-')
            columns_to_sizing[header] = max(len(str(value)), columns_to_sizing[header])

    for header in headers:
        column_size = max(columns_to_sizing[header], len(header)) + 1
        format_string += '{' + header + ':<' + str(column_size) + '}\t'

    print(format_string.format(**{key: key for key in headers}), file=sys.stderr)

    for row in rows:
        defaulted_row = {header: row.get(header) or '-' for header in headers}
        print(format_string.format(**defaulted_row))


def convert_dict_to_tags(tags):
    """Convenience function for converting a dict to datadog tags"""
    return ["{0}:{1}".format(key, value) for key, value in tags.iteritems()]


def convert_tags_to_dict(tags):
    """Convenience function for converting datadog tags to a dict"""
    return dict(tag.split(':', 1) for tag in tags)


def is_truthy(var):
    """Convenience function for checking whether a variable is truthy"""
    if isinstance(var, basestring):
        return var.lower() in YES_LIST
    else:
        return bool(var)



"""
Data Kennel class for orchestrating management of Datadog monitors.
"""
import logging
import json
import difflib

from datadog import api, initialize

from data_kennel.config import MonitorType
from data_kennel.util import convert_dict_to_tags

logger = logging.getLogger(__name__)


class Monitor(object):
    """
    Class for orchestrating management of Datadog monitors.
    """

    def __init__(self, config=None):
        self.real_monitors = {}
        self.config = config

        initialize(
            api_key=self.config.api_key,
            app_key=self.config.app_key
        )

    def list(self, tags=None):
        """
        Returns a list of dictionarys that forms a human-readable table of monitors created by Data Kennel,
        with optional tags for filtering.
        """
        monitors = self.get_monitors(tags)
        printable_monitors = []

        for monitor in monitors:
            if self._is_principal_monitor(monitor):
                name = monitor['name']
                tags = ", ".join(monitor['tags'])
                state = monitor['overall_state']

                printable_monitors.append(
                    {
                        "Name": name,
                        "State": state,
                        "Tags": tags
                    }
                )

        return printable_monitors

    def update(self, dry_run=False, tags=None):
        """
        Orchestrates creation and updating of monitors. If a configured monitor already exists, it is updated
        in place. If it doesn't exist, then it is created. If an existing monitor has no counterpart in the
        configuration, it is deleted.

        dry_run If True, no changes are written to Datadog.
        tags    A dictionary of tags to filter monitors by.
        """
        logger.info('Updating monitors')

        if dry_run:
            logger.info('--dry-run active, no changes will be made')

        configured_monitors = self.config.get_monitors(tags)
        self.real_monitors = self.get_monitors(tags)

        for configured_monitor in configured_monitors:
            # process sub monitors if the monitor is a composite monitor
            sub_monitors = self.config.get_sub_monitor(configured_monitor)
            sub_monitor_ids = []
            for sub_monitor in sub_monitors:
                monitor = self._create_or_update_monitor(sub_monitor, dry_run)
                sub_monitor_ids.append(monitor['id'])

            if sub_monitor_ids:
                # The monitor is a composite monitor. Build the query using the sub_monitor ids
                configured_monitor['query'] = ' && '.join(sub_monitor_ids)

            # Process the principal monitor
            self._create_or_update_monitor(configured_monitor, dry_run)

        # For all of the real monitors that didn't have a configured equivalent, delete them.
        for monitor in self.real_monitors:
            logger.info('Deleting monitor: %s', monitor['name'])

            if not dry_run:
                api.Monitor.delete(monitor['id'])

    def delete(self, dry_run=False, tags=None):
        """
        Deletes monitors.

        dry_run If True, no changes are written to Datadog.
        tags    A dictionary of tags to filter monitors by.
        """
        logger.info('Deleting monitors')

        if dry_run:
            logger.info('--dry-run active, no changes will be made')

        monitors = self.get_monitors(tags)

        for monitor in monitors:
            if self._is_principal_monitor(monitor):
                # Process the principal monitor, The sub_monitor will be handle if the monitor is a composite
                logger.info('Deleting monitor: %s', monitor['name'])
                # Try getting any sub_monitors associated to the principal monitor
                sub_monitors = self._get_sub_monitors(monitor)
                if sub_monitors:
                    logger.info('Deleting sub-monitors: %s', [sub_monitor['name'] for sub_monitor in
                                                              sub_monitors])
                if not dry_run:
                    # delete the principal monitor and  any associated sub_monitors
                    api.Monitor.delete(monitor['id'])
                    for sub_monitor_id in [sub_monitor['id'] for sub_monitor in sub_monitors]:
                        api.Monitor.delete(sub_monitor_id)

    def get_monitors(self, tags=None):
        """
        Gets all existing Datadog monitors, with some convenient filtering.

        tags    A dictionary of tags to filter monitors by.
        """
        monitors = []
        # get monitors for each team that we have a config file for
        for team in self.config.teams:
            default_tags = {'source': 'data_kennel', 'team': team}
            default_tags.update(tags or {})
            monitor_tags = convert_dict_to_tags(default_tags)

            monitors.extend(api.Monitor.get_all(monitor_tags=monitor_tags))

        # Annoyingly, the Datadog API treats monitor tags as ORs instead of ANDs, so we need to do some of our
        # own filtering to achieve that behavior.
        # only do the filtering with the actual tags that the user used (exclude the auto team tag
        # since there can be multiple teams)
        user_tags = convert_dict_to_tags(tags or {})
        return [
            monitor for monitor in monitors
            if all(tag in monitor.get('tags', []) for tag in user_tags)
        ]

    def _compare_monitor(self, monitor1, monitor2):
        """
        Convenience method for comparing two monitors.

        If two monitors have the same name or query, they are considered to be equivalent.
        """
        return monitor1['name'] == monitor2['name'] or monitor1['query'] == monitor2['query']

    def _merge_monitor(self, base_monitor, new_monitor):
        """
        Convenience method for merging two monitors.

        The base monitor is updated with the new monitor's configuration.
        """
        monitor = base_monitor.copy()
        monitor.update(new_monitor)

        base_options = base_monitor.get('options', {}).copy()
        base_options.update(new_monitor.get('options', {}))
        monitor['options'] = base_options

        return monitor

    def _diff_monitors(self, monitor1, monitor1_name, monitor2, monitor2_name):
        """
        Convenience method for performing a diff of two monitors, returning the diff as a string.
        """
        return "\n".join(difflib.unified_diff(
            json.dumps(monitor1, indent=4, sort_keys=True).splitlines(),
            json.dumps(monitor2, indent=4, sort_keys=True).splitlines(),
            fromfile=monitor1_name,
            tofile=monitor2_name
        ))

    def _create_or_update_monitor(self, configured_monitor, dry_run=False):
        """
        Function to create or update a monitor
        :param configured_monitor: The monitor to create or update
        :param dry_run: If True, no changes are written to Datadog.
        :return: the created or updated monitor
        """
        real_monitor = next(
            (monitor for monitor in self.real_monitors if self._compare_monitor(monitor, configured_monitor)),
            None
        )

        # If we found an equivalent real_monitor, we prepare to update it. Otherwise, we'll make a new
        # monitor.
        if real_monitor:
            # Its important to play the configured monitor over the real monitor because the config should
            # be considered the source of truth.
            merged_monitor = self._merge_monitor(real_monitor, configured_monitor)
            # The idea is to eventually make real_monitors the list of all real monitors that don't have
            # equivalents in the config.
            self.real_monitors.remove(real_monitor)

            if merged_monitor != real_monitor:
                logger.info('Updating monitor: %s', merged_monitor['name'])

                monitor_diff = self._diff_monitors(
                    real_monitor,
                    "Existing Monitor",
                    merged_monitor,
                    "New Monitor"
                )

                logger.debug('Differences between monitors:\n%s', monitor_diff)

                if not dry_run:
                    return api.Monitor.update(**merged_monitor)
        else:
            logger.info('Creating monitor: %s', configured_monitor['name'])

            if not dry_run:
                return api.Monitor.create(**configured_monitor)

    def _is_principal_monitor(self, monitor):
        """
        Convenience method for testing if the monitor is a `principal monitor` (not a sub-monitor)
        :param monitor: The monitor
        :return: True if the monitor is a `principal monitor`
        """
        main_monitor_tag = 'dk_type:' + MonitorType.DK_MONITOR
        sub_monitor_tag = 'dk_type:' + MonitorType.DK_SUB_MONITOR

        return main_monitor_tag in monitor['tags'] or sub_monitor_tag not in monitor['tags']

    def _is_composite_monitor(self, monitor):
        """
        Convenience method for testing if the monitor is a `composite monitor`
        :param monitor: The monitor
        :return: True if the monitor is a `composite monitor`
        """
        return ' && ' in monitor.get('query', '')

    def _get_sub_monitors(self, monitor):
        """
        Convenience method for getting the sub-monitor associated to the monitor if the specified monitor
        is a composite monitor
        :param monitor: The monitors
        :return: A list containing the sub-monitors associted to the specified monitor
        """
        sub_monitor_ids = monitor['query'].split('&&') if self._is_composite_monitor(monitor) else []
        sub_monitors = []
        for sub_monitor_id in sub_monitor_ids:
            sub_monitors.append(api.Monitor.get(sub_monitor_id))
        return sub_monitors



"""setup.py controls the build, testing, and distribution of the egg"""

from setuptools import setup, find_packages
import re
import os.path


VERSION_REGEX = re.compile(r"""
    ^__version__\s=\s
    ['"](?P<version>.*?)['"]
""", re.MULTILINE | re.VERBOSE)

VERSION_FILE = os.path.join("data_kennel", "version.py")


def get_long_description():
    """Reads the long description from the README"""
    with open('README.rst') as readmefile:
        return readmefile.read()


def get_version():
    """Reads the version from the package"""
    with open(VERSION_FILE) as handle:
        lines = handle.read()
        version_result = VERSION_REGEX.search(lines)
        if version_result:
            return "{0}".format(version_result.group(1))
        else:
            raise ValueError("Unable to determine __version__")


def get_requirements():
    """Reads the installation requirements from requirements.pip"""
    with open("requirements.pip") as reqfile:
        return filter(lambda line: not line.startswith(('#', '-')), reqfile.read().split("\n"))

setup(
    name='data_kennel',
    version=get_version(),
    description="Data Kennel is a CLI tool for managing Datadog infrastructure.",
    long_description=get_long_description(),
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=[
        'python',
        'datadog',
        'cli'
    ],
    author='amplify',
    author_email='astronauts.core@amplify.com',
    url='https://github.com/amplify-education/data_kennel',
    license='MIT',
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    zip_safe=False,
    dependency_links=[
    ],
    install_requires=get_requirements(),
    scripts=['bin/dk_monitor'],
    test_suite='nose.collector',
    entry_points="""
        [paste.app_factory]
        main=data_kennel:main
    """,
)



"""
Class for parsing Data Kennel's configuration file.
"""
import copy
import glob

import yaml
import os
import re
import json
import logging
from schema import Schema, Optional, Or, Use, SchemaError

from data_kennel.util import convert_dict_to_tags, is_truthy, convert_tags_to_dict
from data_kennel import __version__

DEFAULT_RECOVERY_MESSAGE = "This alert has recovered."
VARIABLE_PATTERN = "(\\$\\{.+?\\})"
SUB_MONITOR_NAME_TEMPLATE = '[DK-C] {0} -- {1}'

CONFIG_SCHEMA = Schema(
    {
        'data_kennel': {
            'team': str
        },
        'monitors': [
            {
                'name': str,
                'query': str,
                'type': Or('metric alert', 'service check', 'event alert', 'query alert'),
                'message': str,
                'notify': [
                    str
                ],
                Optional('tags'): {
                    str: Use(str)
                },
                Optional('with_variables'): [
                    {
                        str: Use(str)
                    }
                ],
                Optional('options'): {
                    Optional('silenced', default=None): {
                        str: Or(None, Use(int))
                    },
                    Optional('notify_no_data'): Use(is_truthy),
                    Optional('new_host_delay'): Or(Use(int), str),
                    Optional('no_data_timeframe'): Or(Use(int), str),
                    Optional('timeout_h'): Use(int),
                    Optional('require_full_window'): Use(is_truthy),
                    Optional('renotify_interval'): Or(Use(int), str),
                    Optional('escalation_message'): str,
                    Optional('notify_audit'): Use(is_truthy),
                    Optional('locked'): Use(is_truthy),
                    Optional('include_tags'): Use(is_truthy),
                    Optional('thresholds'): {
                        Optional('critical'): Or(Use(float), str),
                        Optional('warning'): Or(Use(float), str),
                        Optional('ok'): Or(Use(float), str)
                    },
                    Optional('evaluation_delay'): Use(int),
                    Optional(str): Use(str)
                }
            }
        ]
    }
)

logger = logging.getLogger(__name__)


class MonitorType(object):
    """Enum Class representing the data kennel monitor types"""
    DK_MONITOR = 'Monitor'
    DK_SUB_MONITOR = 'Sub Monitor'


class Config(object):
    """Class for parsing Data Kennel's configuration file."""

    def __init__(self, config_list=None, config_path=None, config_dir=None, api_key=None, app_key=None):
        configs = {}
        self.team_config = {}
        if config_list:
            for config in config_list:
                self._validate_config(config)
                team = config['data_kennel']['team']
                if not configs.get(team):
                    configs[team] = copy.deepcopy(config)
                else:
                    configs[team]['monitors'].extend(config['monitors'])
        elif config_path:
            config = yaml.load(open(config_path))
            self._validate_config(config)
            team = config['data_kennel']['team']
            configs[team] = config
        elif config_dir:
            config_files = glob.glob(config_dir + '/*.yml')
            for conf_file in config_files:
                config = yaml.load(open(conf_file))
                try:
                    self._validate_config(config)
                except SchemaError as ex:
                    raise Exception('Invalid schema in %s: %s' % (conf_file, ex))
                team = config['data_kennel']['team']
                if not configs.get(team):
                    configs[team] = copy.deepcopy(config)
                else:
                    configs[team]['monitors'].extend(config['monitors'])

        self._api_key = api_key
        self._app_key = app_key

        for team in configs.keys():
            self.team_config[team] = self._interpolate_config(configs[team])

    @property
    def teams(self):
        """The teams of the config files"""
        return self.team_config.keys()

    @property
    def api_key(self):
        """Datadog API Key"""
        if self._api_key is None:
            api_key = os.getenv('DATADOG_API_KEY')
            if api_key is None:
                raise Exception('Data Kennel relies on environment variable DATADOG_API_KEY')
            self._api_key = api_key
        return self._api_key

    @property
    def app_key(self):
        """Data Kennel APP Key"""
        if self._app_key is None:
            app_key = os.getenv('DATA_KENNEL_APP_KEY')
            if app_key is None:
                raise Exception('Data Kennel relies on environment variable DATA_KENNEL_APP_KEY')
            self._app_key = app_key
        return self._app_key

    def _get_team(self, config):
        """The team of the config file"""
        return config['data_kennel']['team']

    def _get_team_from_monitor_tags(self, monitor):
        """Extract the team name from the monitor tags"""
        dict_tag = convert_tags_to_dict(monitor['tags'])
        return dict_tag['team']

    def _validate_config(self, config):
        """
        Function for validating that the parsed config object is a valid data_kennel config.
        """
        CONFIG_SCHEMA.validate(config)

    def _build_tags(self, dk_type, team, tags=None):
        """
        Function for building the monitor tags
        :param dk_type: The type of the monitor
        :param team: The Data Kennel Team name
        :param tags: the tags provided in the config file for the monitor
        """
        # Make sure that our default tags are set
        default_tags = {'source': 'data_kennel',
                        'team': team,
                        'dk_version': __version__,
                        'dk_type': dk_type}
        tags = tags or {}
        tags.update(default_tags)
        return tags

    def _interpolate_config(self, config):
        """
        Function for interpolating strings in the config object.

        Returns a copy of the config object with strings interpolated and monitors expanded out.
        """
        interpolated_config = {
            'data_kennel': config['data_kennel'].copy(),
            'monitors': []
        }

        for monitor in config['monitors']:
            variable_sets = monitor.get('with_variables', [{"team": self._get_team(config)}])

            for variable_set in variable_sets:
                variable_set['team'] = self._get_team(config)

                interpolated_monitor = monitor.copy()
                interpolated_monitor.pop('with_variables', None)

                reps = {re.escape('${{{0}}}'.format(str(k))): str(v) for k, v in variable_set.iteritems()}
                pattern = re.compile("|".join(reps.keys()))

                # Dumps the complex dictionary into a simple string. Then replace all of our variables with
                # their actual values there.
                interpolated_monitor_string = pattern.sub(
                    lambda m: reps[re.escape(m.group(0))],
                    json.dumps(interpolated_monitor)
                )

                # If any variables still remain, warn the user but continue.
                matches = re.search(VARIABLE_PATTERN, interpolated_monitor_string)
                if matches is not None:
                    logger.warning(
                        "Non-interpolated variables '%s' found for monitor '%s'",
                        ", ".join(matches.groups()), monitor['name']
                    )
                    continue

                # Load our complex dictionary object back from its string...
                interpolated_monitor = json.loads(interpolated_monitor_string)

                # Build and set the monitor's tags
                interpolated_monitor['tags'] = self._build_tags(MonitorType.DK_MONITOR,
                                                                self._get_team(config),
                                                                interpolated_monitor.get('tags', {}))

                interpolated_config['monitors'].append(interpolated_monitor)

        return interpolated_config

    def get_monitors(self, tags=None):
        """
        Gets monitors in an appropriate format for the Datadog API.

        tags    A dictionary of tags to filter the monitors by.
        """
        configured_monitors = []

        # Tags should really at least be a dict object...
        tags = tags or {}
        # logger.info("team config %s", self.team_config)
        for team in self.teams:
            for monitor in self.team_config[team]['monitors']:
                monitor_tags = monitor.get('tags', {})

                # Test if the requested tags are a subset of the monitor's tags
                if tags.viewitems() <= monitor_tags.viewitems():
                    # Convert tags into the format expected by Datadog's API
                    monitor['tags'] = convert_dict_to_tags(monitor['tags'])

                    # Prefix monitor names with the team name for namespacing
                    monitor['name'] = "[DK] {0} | {1}".format(team, monitor['name'])

                    # Add default recovery message and notificaiton options
                    notifications = " ".join(['@' + notification for notification in monitor['notify']])
                    format_string = (
                        "{{{{#is_alert}}}}\n{0}\n{{{{/is_alert}}}}\n"
                        "{{{{#is_recovery}}}}\n{1}\n{{{{/is_recovery}}}}\n"
                        "{2}"
                    )
                    message = format_string.format(monitor['message'],
                                                   DEFAULT_RECOVERY_MESSAGE,
                                                   notifications)
                    monitor['message'] = message

                    # This isn't needed anymore
                    del monitor['notify']

                    configured_monitors.append(monitor)

        return configured_monitors

    def get_sub_monitor(self, monitor):
        """
        Extract the sub-monitors from the specified monitor in the Datadog format if any exist.
        If the monitor query is a string containing '&&' operator substrings then the monitor is a
        composite monitor and the query elements are the sub-monitor's conditions used by the composite
        monitor.
        :param monitor: The monitor to process
        """
        if isinstance(monitor['query'], basestring) and '&&' not in monitor['query']:
            # The query string does not have any '&&' sub-string, This is a simple monitor
            return {}

        if not isinstance(monitor['query'], basestring):
            # This current version assumes that the query is a string. In the future we will extend Data
            # Kennel to use a list for the query to represent '||' operation
            raise Exception('List in monitor query is not yet supported.')

        sub_monitors = []
        sub_queries = monitor['query'].split('&&')
        index = 1
        name = monitor['name'][len('[DK] '):] if monitor['name'].startswith('[DK] ') else monitor['name']
        tags = convert_dict_to_tags(self._build_tags(dk_type=MonitorType.DK_SUB_MONITOR,
                                                     team=self._get_team_from_monitor_tags(monitor)))

        for query in sub_queries:
            # Create a new monitor for each
            sub_monitor = {'name': SUB_MONITOR_NAME_TEMPLATE.format(name, index),
                           'type': monitor['type'],
                           'tags': tags,
                           }
            if monitor.get('options'):
                sub_monitor['options'] = monitor.get('options')
            sub_monitor['query'] = query.strip()
            sub_monitors.append(sub_monitor)
            index += 1

        return sub_monitors
